import magicattr

from dspy.primitives.assertions import *
from dspy.primitives.module import BaseModule


class ProgramMeta(type):
    pass
    # def __call__(cls, *args, **kwargs):
    #     obj = super(ProgramMeta, cls).__call__(*args, **kwargs)

    #     if issubclass(cls, Program) and not getattr(obj, "_program_init_called", False):
    #         obj._base_init()
    #         obj._program_init_called = True
    #     return obj


class Module(BaseModule, metaclass=ProgramMeta):
    def _base_init(self):
        self._compiled = False

    def __init__(self):
        self._compiled = False

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def named_predictors(self):
        from dspy.predict.predict import Predict

        return [(name, param) for name, param in self.named_parameters() if isinstance(param, Predict)]

    def predictors(self):
        return [param for _, param in self.named_predictors()]
    
    def set_lm(self, lm):
        import dspy
        assert dspy.settings.experimental, "Setting the lm is an experimental feature."

        for _, param in self.named_predictors():
            param.lm = lm

    def get_lm(self):
        import dspy
        assert dspy.settings.experimental, "Getting the lm is an experimental feature."

        all_used_lms = [param.lm for _, param in self.named_predictors()]

        if len(set(all_used_lms)) == 1:
            return all_used_lms[0]
        
        raise ValueError("Multiple LMs are being used in the module.")

    def __repr__(self):
        s = []

        for name, param in self.named_predictors():
            s.append(f"{name} = {param}")

        return "\n".join(s)

    def map_named_predictors(self, func):
        """Applies a function to all named predictors."""
        for name, predictor in self.named_predictors():
            set_attribute_by_name(self, name, func(predictor))
        return self

    def activate_assertions(self, handler=backtrack_handler, **handler_args):
        """
        Activates assertions for the module.
        The default handler is the backtrack_handler.
        """
        assert_transform_module(self, handler, **handler_args)
        return self

    # def __deepcopy__(self, memo):
    #     # memo is a dict of id's to copies already made during the current call
    #     # Check if the object is already copied
    #     if id(self) in memo:
    #         return memo[id(self)]

    #     print(f"Deep copying {self.__class__.__name__}...")

    #     new_copy = copy.copy(self)
    #     memo[id(self)] = new_copy

    #     for k, v in self.__dict__.items():
    #         print(f"Copying attribute {k} of type {type(v)}...")
    #         setattr(new_copy, k, copy.deepcopy(v, memo))
    #         print("Done")

    #     return new_copy


def set_attribute_by_name(obj, name, value):
    magicattr.set(obj, name, value)

Program = Module


class SentenceTransformersCrossEncoder:
    """Wrapper for sentence-transformers cross-encoder model.
    """
    def __init__(
        self, model_name_or_path: str = "cross-encoder/ms-marco-MiniLM-L-12-v2",
    ):
        try:
            from sentence_transformers.cross_encoder import CrossEncoder
        except ImportError:
            raise ModuleNotFoundError(
                "You need to install sentence-transformers library to use SentenceTransformersCrossEncoder.",
            )
        self.model = CrossEncoder(model_name_or_path)

    def __call__(self, query: str, passage: list[str]) -> list[float]:
        return self.model.predict([[query, p] for p in passage]).tolist()



import os
import tqdm
import copy
import datetime
import itertools

from collections import defaultdict

def print_message(*s, condition=True, pad=False, sep=None):
    s = " ".join([str(x) for x in s])
    msg = "[{}] {}".format(datetime.datetime.now().strftime("%b %d, %H:%M:%S"), s)

    if condition:
        msg = msg if not pad else f"\n{msg}\n"
        print(msg, flush=True, sep=sep)

    return msg


def timestamp(daydir=False):
    format_str = f"%Y-%m{'/' if daydir else '-'}%d{'/' if daydir else '_'}%H.%M.%S"
    result = datetime.datetime.now().strftime(format_str)
    return result


def file_tqdm(file):
    print(f"#> Reading {file.name}")

    with tqdm.tqdm(
        total=os.path.getsize(file.name) / 1024.0 / 1024.0, unit="MiB",
    ) as pbar:
        for line in file:
            yield line
            pbar.update(len(line) / 1024.0 / 1024.0)

        pbar.close()


def create_directory(path):
    if os.path.exists(path):
        print("\n")
        print_message("#> Note: Output directory", path, "already exists\n\n")
    else:
        print("\n")
        print_message("#> Creating directory", path, "\n\n")
        os.makedirs(path)


def deduplicate(seq: list[str]) -> list[str]:
    """
    Source: https://stackoverflow.com/a/480227/1493011
    """

    seen = set()
    return [x for x in seq if not (x in seen or seen.add(x))]


def batch(group, bsize, provide_offset=False):
    offset = 0
    while offset < len(group):
        L = group[offset : offset + bsize]
        yield ((offset, L) if provide_offset else L)
        offset += len(L)
    return


# class dotdict(dict):
#     """
#     dot.notation access to dictionary attributes
#     Credit: derek73 @ https://stackoverflow.com/questions/2352181
#     """

#     __getattr__ = dict.__getitem__
#     __setattr__ = dict.__setitem__
#     __delattr__ = dict.__delitem__


class dotdict(dict):
    def __getattr__(self, key):
        if key.startswith('__') and key.endswith('__'):
            return super().__getattr__(key)
        try:
            return self[key]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def __setattr__(self, key, value):
        if key.startswith('__') and key.endswith('__'):
            super().__setattr__(key, value)
        else:
            self[key] = value

    def __delattr__(self, key):
        if key.startswith('__') and key.endswith('__'):
            super().__delattr__(key)
        else:
            del self[key]

    def __deepcopy__(self, memo):
        # Use the default dict copying method to avoid infinite recursion.
        return dotdict(copy.deepcopy(dict(self), memo))


class dotdict_lax(dict):
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def flatten(L):
    # return [x for y in L for x in y]

    result = []
    for _list in L:
        result += _list

    return result


def zipstar(L, lazy=False):
    """
    A much faster A, B, C = zip(*[(a, b, c), (a, b, c), ...])
    May return lists or tuples.
    """

    if len(L) == 0:
        return L

    width = len(L[0])

    if width < 100:
        return [[elem[idx] for elem in L] for idx in range(width)]

    L = zip(*L)

    return L if lazy else list(L)


def zip_first(L1, L2):
    length = len(L1) if type(L1) in [tuple, list] else None

    L3 = list(zip(L1, L2))

    assert length in [None, len(L3)], "zip_first() failure: length differs!"

    return L3


def int_or_float(val):
    if "." in val:
        return float(val)

    return int(val)


def groupby_first_item(lst):
    groups = defaultdict(list)

    for first, *rest in lst:
        rest = rest[0] if len(rest) == 1 else rest
        groups[first].append(rest)

    return groups


def process_grouped_by_first_item(lst):
    """
    Requires items in list to already be grouped by first item.
    """

    groups = defaultdict(list)

    started = False
    last_group = None

    for first, *rest in lst:
        rest = rest[0] if len(rest) == 1 else rest

        if started and first != last_group:
            yield (last_group, groups[last_group])
            assert (
                first not in groups
            ), f"{first} seen earlier --- violates precondition."

        groups[first].append(rest)

        last_group = first
        started = True

    return groups


def grouper(iterable, n, fillvalue=None):
    """
    Collect data into fixed-length chunks or blocks
        Example: grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
        Source: https://docs.python.org/3/library/itertools.html#itertools-recipes
    """

    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def lengths2offsets(lengths):
    offset = 0

    for length in lengths:
        yield (offset, offset + length)
        offset += length

    return


# see https://stackoverflow.com/a/45187287
class NullContextManager:
    def __init__(self, dummy_resource=None):
        self.dummy_resource = dummy_resource

    def __enter__(self):
        return self.dummy_resource

    def __exit__(self, *args):
        pass


def load_batch_backgrounds(args, qids):
    if args.qid2backgrounds is None:
        return None

    qbackgrounds = []

    for qid in qids:
        back = args.qid2backgrounds[qid]

        if len(back) and type(back[0]) == int:
            x = [args.collection[pid] for pid in back]
        else:
            x = [args.collectionX.get(pid, "") for pid in back]

        x = " [SEP] ".join(x)
        qbackgrounds.append(x)

    return qbackgrounds



"""Module for interacting with Snowflake Cortex."""
import json
from typing import Any

import backoff

from dsp.modules.lm import LM
from dsp.utils.settings import settings

try:
    from snowflake.snowpark import functions as snow_func
except ImportError:
    pass


def backoff_hdlr(details) -> None:
    """Handler from https://pypi.org/project/backoff ."""
    print(
        f"Backing off {details['wait']:0.1f} seconds after {details['tries']} tries ",
        f"calling function {details['target']} with kwargs",
        f"{details['kwargs']}",
    )


def giveup_hdlr(details) -> bool:
    """Wrapper function that decides when to give up on retry."""
    if "rate limits" in str(details):
        return False
    return True


class Snowflake(LM):
    """Wrapper around Snowflake's CortexAPI.

    Supported models include 'llama3.1-70b','llama3.1-405b','snowflake-arctic','mistral-large','reka-flash','mixtral-8x7b',
    'llama2-70b-chat','mistral-7b','gemma-7b','llama3-8b','llama3-70b','reka-core'.
    """

    def __init__(self, session: object, model: str = "mixtral-8x7b", **kwargs):
        """Parameters

        ----------
        session:
            Snowflake Snowpark session for accessing Snowflake Cortex service.
            Full list of requirements can be found here: https://docs.snowflake.com/en/developer-guide/snowpark/reference/python/latest/api/snowflake.snowpark.Session
        model : str
            Which pre-trained model from Snowflake to use.
            Choices are 'snowflake-arctic','mistral-large','reka-flash','mixtral-8x7b','llama2-70b-chat','mistral-7b','gemma-7b'
            Full list of supported models is available here: https://docs.snowflake.com/en/user-guide/snowflake-cortex/llm-functions#complete
        **kwargs: dict
            Additional arguments to pass to the API provider.
        """
        super().__init__(model)

        self.client = self._init_cortex(snowflake_session=session)
        self.model = model
        self.available_args = {
            "max_tokens",
            "temperature",
            "top_p",
        }

        self.provider = "Snowflake"
        self.history: list[dict[str, Any]] = []
        self.kwargs = {
            **self.kwargs,
            "temperature": 0.7,
            "max_output_tokens": 1024,
            "top_p": 1.0,
            "top_k": 1,
            **kwargs,
        }

    @classmethod
    def _init_cortex(cls, snowflake_session) -> None:
        # session = Session.builder.configs(credentials).create()
        snowflake_session.query_tag = {"origin": "sf_sit", "name": "dspy", "version": {"major": 1, "minor": 0}}

        return snowflake_session

    def _prepare_params(
        self,
        parameters: Any,
    ) -> dict:
        params_mapping = {"n": "candidate_count", "max_tokens": "max_output_tokens"}
        params = {params_mapping.get(k, k): v for k, v in parameters.items()}
        params = {**self.kwargs, **params}
        return {k: params[k] for k in set(params.keys()) & self.available_args}

    def _cortex_complete_request(self, prompt: str, **kwargs) -> dict:
        complete = snow_func.builtin("snowflake.cortex.complete")
        cortex_complete_args = complete(
            snow_func.lit(self.model),
            snow_func.lit([{"role": "user", "content": prompt}]),
            snow_func.lit(kwargs),
        )
        raw_response = self.client.range(1).withColumn("complete_cal", cortex_complete_args).collect()

        if len(raw_response) > 0:
            return json.loads(raw_response[0].COMPLETE_CAL)

        else:
            return json.loads('{"choices": [{"messages": "None"}]}')

    def basic_request(self, prompt: str, **kwargs) -> list:
        raw_kwargs = kwargs
        kwargs = self._prepare_params(raw_kwargs)

        response = self._cortex_complete_request(prompt, **kwargs)

        history = {
            "prompt": prompt,
            "response": {
                "prompt": prompt,
                "choices": [{"text": c} for c in response["choices"]],
            },
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }

        self.history.append(history)

        return [i["text"]["messages"] for i in history["response"]["choices"]]

    @backoff.on_exception(
        backoff.expo,
        (Exception),
        max_time=settings.backoff_time,
        on_backoff=backoff_hdlr,
        giveup=giveup_hdlr,
    )
    def _request(self, prompt: str, **kwargs):
        """Handles retrieval of completions from Snowflake Cortex whilst handling API errors."""
        return self.basic_request(prompt, **kwargs)

    def __call__(
        self,
        prompt: str,
        only_completed: bool = True,
        return_sorted: bool = False,
        **kwargs,
    ):
        return self._request(prompt, **kwargs)



# import os
# import random
# import subprocess
# import time

# import tqdm
# import ujson
# from datasets.fingerprint import Hasher

# import dsp

# if os.environ.get('DSP_NOTEBOOK_CACHEDIR'):
#     training_data_directory = os.path.join(os.environ.get('DSP_NOTEBOOK_CACHEDIR'), 'compiler')
# else:
#     training_data_directory = 'cache/compiler'


# compilations_assumed_to_exist={'ft-zvEdzQVQ5xwlxvNPrxl6kpnw': 'ada:ft-stanfordpraglab-2023-02-09-19-50-49'}


# def openai_check_finetune(jobname):
#     if dsp.settings.force_reuse_cached_compilation and jobname in compilations_assumed_to_exist:
#         return compilations_assumed_to_exist[jobname]

#     command = f"""openai api fine_tunes.get -i {jobname}"""
#     print(command)

#     result = subprocess.run(command.split(), stdout=subprocess.PIPE, check=False)
#     output = result.stdout.decode("utf-8").strip()

#     try:
#         output = ujson.loads(output)
#         if output['status'] == 'succeeded':
#             return output['fine_tuned_model']

#         if output['status'] in ['pending', 'running']:
#             print(f'Compiling, run ```openai api fine_tunes.follow -i {jobname}``` for details...')
#             time.sleep(60)
#             return openai_check_finetune(jobname)
#     except:
#         pass

#     return False


# def convert_to_training_point2(y, inputs, outputs, template):
#     assert len(inputs) + len(outputs) == len(template.fields)

#     y_ = dsp.Example(**{f: y[f] for f in inputs}, demos=[])
#     prompt = template(y_, show_guidelines=False)

#     completion = y[outputs[0]]
#     output_fields = template.fields[len(inputs):]

#     for field in output_fields[1:]:
#         completion += f"\n\n{field.name} " + y[field.output_variable]
    
#     completion = " " + completion + " </s>"
#     return {'prompt': prompt, 'completion': completion}


# def simulate(program, input_examples):
#     training_data = []

#     for input_example in tqdm.tqdm(input_examples):
#         prediction = program(input_example)

#         if prediction is not None:
#             # assert len(prediction.compiling_stages) == 2, "TMP"
#             for stage in prediction.compiling_stages:
#                 name, template, inputs, outputs = stage['name'], stage['template'], stage['inputs'], stage['outputs']
#                 training_data.append(convert_to_training_point2(prediction.get(name), inputs, outputs, template))
    
#     r = random.Random(0)
#     r.shuffle(training_data)

#     return training_data


# def openai_finetune_(name, target):
#     training_data_path = name_to_path(name)

#     # Launch the fine-tune on the path
#     command = f"""openai api fine_tunes.create -t {training_data_path} -m {target} --n_epochs 4 --learning_rate_multiplier 0.05 --no_check_if_files_exist"""
#     print(command)

#     # command = """python script.py"""
#     process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

#     while line := process.stdout.readline().decode().strip():
#         if 'created fine-tune:' in line.lower():
#             jobname = line.split()[-1]
#             break
        
#     #     if 'costs $' in line.lower():
#     #         cost = line.split()[-1]
#     #         break

#     # assert cost[0] == '$'
    
#     # if float(cost[1:]) > 300:
#     #     print(f'Got cost {cost} -- you may wanna cancel the job: openai api fine_tunes.cancel -i {jobname}')

#     # print(cost)

#     print(jobname)

#     # Block until it's done
#     ft = openai_check_finetune(jobname)
#     assert ft, ft

#     # Return its name
#     return (jobname, ft)


# def openai_finetune(name, target):
#     print(name)
#     training_data_path = name_to_path(name)
#     training_data_path += '.model'

#     # if path + stuff exists, load the tuple from it
#     try:
#         with open(training_data_path) as f:
#             jobname, ft = ujson.loads(f.readline())

#         if openai_check_finetune(jobname):
#             return jobname, ft
#     except:
#         pass
    
#     jobname, ft = openai_finetune_(name, target)

#     with open(training_data_path, 'w') as f:
#         f.write(ujson.dumps((jobname, ft)) + '\n')
    
#     return jobname, ft


# def name_to_path(name):
#     if not os.path.exists(training_data_directory):
#         os.makedirs(training_data_directory)

#     training_data_path = os.path.join(training_data_directory, f'{name}.jsonl')
#     return training_data_path


# # 3. Check that the output file name has status "success" (not deleted or non-existent). Otherwise, re-call with n = n+1.
# def finetune(training_data, target):
#     name = Hasher.hash(training_data)
#     training_data_path = name_to_path(name)

#     with open(training_data_path, 'w') as f:
#         for line in training_data:
#             f.write(ujson.dumps(line) + '\n')

#     jobname, ft = openai_finetune(name, target)
#     print(ft)

#     ft = dsp.GPT3(model=ft, stop=" </s>")
#     return ft

# # 4. Return updated program.
# def compile(program, examples, target='ada'):
#     training_data = simulate(program, examples)
#     compiled_lm = finetune(training_data, target=target)

#     def compiled_program(*args, **kwargs):
#         with dsp.settings.context(compiled_lm=compiled_lm, compiling=False):
#             return program(*args, **kwargs)

#     compiled_program.lm = compiled_lm
#     return compiled_program




from typing import Any, Optional

from pydantic import BaseModel, model_validator


class SynthesizerArguments(BaseModel):
    feedback_mode: Optional[str] = None
    num_example_for_feedback: Optional[int] = None

    input_lm_model: Optional[Any] = None
    output_lm_model: Optional[Any] = None
    output_teacher_module: Optional[Any] = None

    num_example_for_optim: Optional[int] = None

    @model_validator(mode='after')
    def validate_feedback_mode(self):
        if self.feedback_mode and self.feedback_mode not in ["human", "llm"]:
            raise ValueError("Feedback mode should be either 'human' or 'llm'.")
        
        if self.feedback_mode and not self.num_example_for_feedback:
            raise ValueError("Number of examples for feedback is required when feedback mode is provided.")

        return self


import random
from typing import Dict, List, Optional, Union

import dsp
from dspy.predict.parameter import Parameter
from dspy.primitives.prediction import Prediction


def single_query_passage(passages):
    passages_dict = {key: [] for key in list(passages[0].keys())}
    for docs in passages:
        for key, value in docs.items():
            passages_dict[key].append(value)
    if "long_text" in passages_dict:
        passages_dict["passages"] = passages_dict.pop("long_text")
    return Prediction(**passages_dict)


class Retrieve(Parameter):
    name = "Search"
    input_variable = "query"
    desc = "takes a search query and returns one or more potentially relevant passages from a corpus"

    def __init__(self, k=3):
        self.stage = random.randbytes(8).hex()
        self.k = k

    def reset(self):
        pass

    def dump_state(self, save_verbose=False):
        """save_verbose is set as a default argument to support the inherited Parameter interface for dump_state"""
        state_keys = ["k"]
        return {k: getattr(self, k) for k in state_keys}

    def load_state(self, state):
        for name, value in state.items():
            setattr(self, name, value)

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def forward(
        self,
        query_or_queries: Union[str, List[str]],
        k: Optional[int] = None,
        by_prob: bool = True,
        with_metadata: bool = False,
        **kwargs,
    ) -> Union[List[str], Prediction, List[Prediction]]:
        # queries = [query_or_queries] if isinstance(query_or_queries, str) else query_or_queries
        # queries = [query.strip().split('\n')[0].strip() for query in queries]

        # # print(queries)
        # # TODO: Consider removing any quote-like markers that surround the query too.
        # k = k if k is not None else self.k
        # passages = dsp.retrieveEnsemble(queries, k=k,**kwargs)
        # return Prediction(passages=passages)
        queries = (
            [query_or_queries]
            if isinstance(query_or_queries, str)
            else query_or_queries
        )
        queries = [query.strip().split("\n")[0].strip() for query in queries]

        # print(queries)
        # TODO: Consider removing any quote-like markers that surround the query too.
        k = k if k is not None else self.k
        if not with_metadata:
            passages = dsp.retrieveEnsemble(queries, k=k, by_prob=by_prob, **kwargs)
            return Prediction(passages=passages)
        else:
            passages = dsp.retrieveEnsemblewithMetadata(
                queries, k=k, by_prob=by_prob, **kwargs,
            )
            if isinstance(passages[0], List):
                pred_returns = []
                for query_passages in passages:
                    passages_dict = {
                        key: []
                        for key in list(query_passages[0].keys())
                        if key != "tracking_idx"
                    }
                    for psg in query_passages:
                        for key, value in psg.items():
                            if key == "tracking_idx":
                                continue
                            passages_dict[key].append(value)
                    if "long_text" in passages_dict:
                        passages_dict["passages"] = passages_dict.pop("long_text")
                    pred_returns.append(Prediction(**passages_dict))
                return pred_returns
            elif isinstance(passages[0], Dict):
                # passages dict will contain {"long_text":long_text_list,"metadatas";metadatas_list...}
                return single_query_passage(passages=passages)


# TODO: Consider doing Prediction.from_completions with the individual sets of passages (per query) too.


class RetrieveThenRerank(Parameter):
    name = "Search"
    input_variable = "query"
    desc = "takes a search query and returns one or more potentially relevant passages followed by reranking from a corpus"

    def __init__(self, k=3):
        self.stage = random.randbytes(8).hex()
        self.k = k

    def reset(self):
        pass

    def dump_state(self, save_verbose=False):
        """save_verbose is set as a default argument to support the inherited Parameter interface for dump_state"""
        state_keys = ["k"]
        return {k: getattr(self, k) for k in state_keys}

    def load_state(self, state):
        for name, value in state.items():
            setattr(self, name, value)

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def forward(
        self,
        query_or_queries: Union[str, List[str]],
        k: Optional[int] = None,
        with_metadata: bool = False,
        **kwargs,
    ) -> Union[List[str], Prediction, List[Prediction]]:
        queries = (
            [query_or_queries]
            if isinstance(query_or_queries, str)
            else query_or_queries
        )
        queries = [query.strip().split("\n")[0].strip() for query in queries]

        # print(queries)
        # TODO: Consider removing any quote-like markers that surround the query too.
        k = k if k is not None else self.k
        if not with_metadata:
            passages = dsp.retrieveRerankEnsemble(queries, k=k, **kwargs)
            return passages
        else:
            passages = dsp.retrieveRerankEnsemblewithMetadata(queries, k=k, **kwargs)
            if isinstance(passages[0], List):
                pred_returns = []
                for query_passages in passages:
                    passages_dict = {key: [] for key in list(query_passages[0].keys())}
                    for docs in query_passages:
                        for key, value in docs.items():
                            passages_dict[key].append(value)
                    if "long_text" in passages_dict:
                        passages_dict["passages"] = passages_dict.pop("long_text")

                    pred_returns.append(Prediction(**passages_dict))
                return pred_returns
            elif isinstance(passages[0], Dict):
                return single_query_passage(passages=passages)



from abc import ABC, abstractmethod


class Proposer(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def propose_instructions_for_program(self):
        pass

    def propose_instruction_for_predictor(self):
        pass


from pathlib import Path
from typing import Any, Optional, Union

from dsp.modules.lm import LM

## Utility functions to load models


def load_tensorrt_model(
    engine_dir: Union[str, Path],
    use_py_session: Optional[bool] = False,
    **kwargs,
) -> tuple[Any, dict]:
    import tensorrt_llm
    from tensorrt_llm.runtime import ModelRunner, ModelRunnerCpp

    runtime_rank = tensorrt_llm.mpi_rank()
    runner_cls = ModelRunner if use_py_session else ModelRunnerCpp
    runner_kwargs = {
        "engine_dir": engine_dir,
        "lora_dir": kwargs.get("lora_dir", None),
        "rank": runtime_rank,
        "lora_ckpt_source": kwargs.get("lora_ckpt_source", "hf"),
    }

    if not use_py_session:
        engine_cpp_kwargs = {}
        defaults = {
            "max_batch_size": 1,
            "max_input_len": 1024,
            "max_output_len": 1024,
            "max_beam_width": 1,
            "max_attention_window_size": None,
            "sink_token_length": None,
        }

        for key, value in defaults.items():
            engine_cpp_kwargs[key] = kwargs.get(key, value)
        runner_kwargs.update(**engine_cpp_kwargs)

    runner = runner_cls.from_dir(**runner_kwargs)
    return runner, runner_kwargs


def tokenize(prompt: Union[list[dict], str], tokenizer: Any, **kwargs) -> list[int]:
    defaults = {
        "add_special_tokens": False,
        "max_input_length": 1024,
        "model_name": None,
        "model_version": None,
    }
    if not isinstance(prompt, str):
        prompt = tokenizer.apply_chat_template(prompt, tokenize=False)

    input_ids = [
        tokenizer.encode(
            prompt,
            add_special_tokens=kwargs.get("add_special_tokens", defaults["add_special_tokens"]),
            truncation=True,
            max_length=kwargs.get("max_input_length", defaults["max_input_length"]),
        ),
    ]
    if (
        kwargs.get("model_name", defaults["model_name"]) == "ChatGLMForCausalLM"
        and kwargs.get("model_version", defaults["model_version"]) == "glm"
    ):
        input_ids.append(tokenizer.stop_token_id)
    return input_ids


class TensorRTModel(LM):
    """TensorRT integration for dspy LM."""

    def __init__(self, model_name_or_path: str, engine_dir: str, **engine_kwargs: dict) -> None:
        """Initialize the TensorRTModel.

        Args:
            model_name_or_path (str): The Huggingface ID or the path where tokenizer files exist.
            engine_dir (str): The folder where the TensorRT .engine file exists.
            **engine_kwargs (Optional[dict]): Additional engine loading keyword arguments.

        Keyword Args:
            use_py_session (bool, optional): Whether to use a Python session or not. Defaults to False.
            lora_dir (str): The directory of LoRA adapter weights.
            lora_task_uids (list[str]): list of LoRA task UIDs; use -1 to disable the LoRA module.
            lora_ckpt_source (str): The source of the LoRA checkpoint.

            If use_py_session is set to False, the following kwargs are supported:
                max_batch_size (int, optional): The maximum batch size. Defaults to 1.
                max_input_len (int, optional): The maximum input context length. Defaults to 1024.
                max_output_len (int, optional): The maximum output context length. Defaults to 1024.
                max_beam_width (int, optional): The maximum beam width, similar to `n` in OpenAI API. Defaults to 1.
                max_attention_window_size (int, optional): The attention window size that controls the
                    sliding window attention / cyclic KV cache behavior. Defaults to None.
                sink_token_length (int, optional): The sink token length. Defaults to 1.
        """
        # Implementation here
        self.model_name_or_path, self.engine_dir = model_name_or_path, engine_dir
        super().__init__(model=self.model_name_or_path)
        try:
            import tensorrt_llm
        except ImportError as exc:
            raise ModuleNotFoundError(
                "You need to install tensorrt-llm to use TensorRTModel",
            ) from exc

        try:
            from transformers import AutoTokenizer
        except ImportError as exc:
            raise ModuleNotFoundError(
                "You need to install torch and transformers ",
                "pip install transformers==4.38.2",
            ) from exc

        # Configure tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_name_or_path,
            legacy=False,
            padding_side="left",
            truncation_side="left",
            trust_remote_code=True,
            use_fast=True,
        )

        self.pad_id = (
            self.tokenizer.eos_token_id if self.tokenizer.pad_token_id is None else self.tokenizer.pad_token_id
        )
        self.end_id = self.tokenizer.eos_token_id

        # Configure TensorRT
        self.runtime_rank = tensorrt_llm.mpi_rank()
        self.runner, self._runner_kwargs = load_tensorrt_model(engine_dir=self.engine_dir, **engine_kwargs)
        self.history: list[dict[str, Any]] = []

    def _generate(self, prompt: Union[list[dict[str, str]], str], **kwargs: dict) -> tuple[list[str], dict]:
        import torch

        input_ids = tokenize(prompt=prompt, tokenizer=self.tokenizer, **kwargs)
        input_ids = torch.tensor(input_ids, dtype=torch.int32)

        run_kwargs = {}
        defaults = {
            "max_new_tokens": 1024,
            "max_attention_window_size": None,
            "sink_token_length": None,
            "end_id": self.end_id,
            "pad_id": self.pad_id,
            "temperature": 1.0,
            "top_k": 1,
            "top_p": 0.0,
            "num_beams": 1,
            "length_penalty": 1.0,
            "early_stopping": 1,
            "repetition_penalty": 1.0,
            "presence_penalty": 0.0,
            "frequency_penalty": 0.0,
            "stop_words_list": None,
            "bad_words_list": None,
            "streaming": False,
            "return_dict": True,
            "output_log_probs": False,
            "output_cum_log_probs": False,
            "output_sequence_lengths": True,
        }

        for k, v in defaults.items():
            run_kwargs[k] = kwargs.get(k, v)

        with torch.no_grad():
            outputs = self.runner.generate(input_ids, **run_kwargs)
        input_lengths = [x.size(0) for x in input_ids]

        output_ids, sequence_lengths = outputs["output_ids"], outputs["sequence_lengths"]

        # In case of current version of dspy it will always stay as 1
        _, num_beams, _ = output_ids.size()
        batch_idx, beams = 0, []

        for beam in range(num_beams):
            output_begin = input_lengths[batch_idx]
            output_end = sequence_lengths[batch_idx][beam]
            outputs = output_ids[batch_idx][beam][output_begin:output_end].tolist()
            output_text = self.tokenizer.decode(outputs)
            beams.append(output_text)

        return beams, run_kwargs

    def basic_request(self, prompt, **kwargs: dict) -> list[str]:
        raw_kwargs = kwargs
        response, all_kwargs = self._generate(prompt, **kwargs)
        history = {
            "prompt": prompt,
            "response": response,
            "raw_kwargs": raw_kwargs,
            "kwargs": all_kwargs,
        }
        self.history.append(history)
        return response

    def __call__(
        self,
        prompt: Union[list[dict[str, str]], str],
        **kwargs,
    ):
        """TensorRTLLM generate method in dspy.

        Args:
            prompt (Union[list[dict[str, str]], str]): The prompt to pass. If prompt is not string
                then it will assume that chat mode / instruct mode is triggered.
            **kwargs (Optional[dict]): Optional keyword arguments.

        Additional Parameters:
            max_new_tokens (int): The maximum number of tokens to output. Defaults to 1024
            max_attention_window_size (int) Defaults to None
            sink_token_length (int): Defaults to None
            end_id (int): The end of sequence of ID of tokenize, defaults to tokenizer's default
                end id
            pad_id (int): The pd sequence of ID of tokenize, defaults to tokenizer's default end id
            temperature (float): The temperature to control probabilistic behaviour in generation
                Defaults to 1.0
            top_k (int): Defaults to 1
            top_p (float): Defaults to 1
            num_beams: (int): The number of responses to generate. Defaults to 1
            length_penalty (float): Defaults to 1.0
            repetition_penalty (float): Defaults to 1.0
            presence_penalty (float): Defaults to 0.0
            frequency_penalty (float): Defaults to 0.0
            early_stopping (int): Use this only when num_beams > 1, Defaults to 1
        """
        return self.request(prompt, **kwargs)



import random
import sys
import textwrap
from collections import defaultdict

import numpy as np
import optuna
import dspy

from dspy.evaluate.evaluate import Evaluate
from dspy.propose import GroundedProposer
from dspy.teleprompt.teleprompt import Teleprompter
from dspy.teleprompt.utils import (
    create_n_fewshot_demo_sets,
    eval_candidate_program,
    get_program_with_highest_avg_score,
    get_signature,
    get_task_model_history_for_full_example,
    print_full_program,
    save_candidate_program,
    set_signature,
)

"""
USAGE SUGGESTIONS:

The following code can be used to compile a optimized signature teleprompter using MIPRO, and evaluate it on an end task:

``` python
from dspy.teleprompt import MIPROv2

teleprompter = MIPROv2(prompt_model=prompt_model, task_model=task_model, metric=metric, num_candidates=10, init_temperature=1.0)
kwargs = dict(num_threads=NUM_THREADS, display_progress=True, display_table=0)
compiled_prompt_opt = teleprompter.compile(program, trainset=trainset[:TRAIN_NUM], num_trials=100, max_bootstrapped_demos=3, max_labeled_demos=5)
eval_score = evaluate(compiled_prompt_opt, devset=valset[:EVAL_NUM], **kwargs)
```

Note that this teleprompter takes in the following parameters:

* prompt_model: The model used for prompt generation. When unspecified, defaults to the model set in settings (i.e., dspy.settings.configure(lm=task_model)).
* task_model: The model used for running your task. When unspecified, defaults to the model set in settings (i.e., dspy.settings.configure(lm=task_model)).
* teacher_settings: The settings used for the teacher model. When unspecified, defaults to the settings set in settings (i.e., dspy.settings.configure(lm=task_model)).
    The teacher settings are used to generate the fewshot examples.  This is the LLM/settings to use as a task model for the bootstrapping runs.
    Typically you would want to use a model of equal or greater quality to your task model.
* metric: The task metric used for optimization.
* num_candidates: The number of new prompts and sets of fewshot examples to generate and evaluate. Default=10.
* init_temperature: The temperature used to generate new prompts. Higher roughly equals more creative. Default=1.0.
* verbose: Tells the method whether or not to print intermediate steps.
* track_stats: Tells the method whether or not to track statistics about the optimization process.
                If True, the method will track a dictionary with a key corresponding to the trial number, 
                and a value containing a dict with the following keys:
                    * program: the program being evaluated at a given trial
                    * score: the last average evaluated score for the program
                    * pruned: whether or not this program was pruned
                This information will be returned as attributes of the best program.
* log_dir: The directory to save logs and other information to. If unspecified, no logs will be saved.
* view_data_batch_size: The number of examples to view in the data batch when producing the dataset summary. Default=10.
* minibatch_size: The size of the minibatch to use when evaluating the program if using minibatched evaluations. Default=25.
* minibatch_full_eval_steps: The number of steps to take before doing a full evaluation of the program if using minibatched evaluations. Default=10.
* metric_threshold: If the metric yields a numerical value, then check it against this threshold when deciding whether or not to accept a bootstrap example.
"""

BOOTSTRAPPED_FEWSHOT_EXAMPLES_IN_CONTEXT = 3
LABELED_FEWSHOT_EXAMPLES_IN_CONTEXT = 0

class MIPROv2(Teleprompter):
    def __init__(
        self,
        metric,
        prompt_model=dspy.settings.lm,
        task_model=dspy.settings.lm,
        teacher_settings={},
        num_candidates=10,
        num_threads=6,
        max_errors=10,
        init_temperature=0.5,
        verbose=False,
        track_stats=True,
        log_dir=None,
        metric_threshold=None,
    ):
        self.num_candidates = num_candidates
        self.metric = metric
        self.init_temperature = init_temperature
        self.task_model = task_model
        self.prompt_model = prompt_model
        self.verbose = verbose
        self.track_stats = track_stats
        self.log_dir = log_dir
        self.teacher_settings = teacher_settings
        self.prompt_model_total_calls = 0
        self.total_calls = 0
        self.num_threads = num_threads
        self.max_errors = max_errors
        self.metric_threshold = metric_threshold

    def compile(
        self,
        student,
        *,
        trainset,
        valset=None,
        num_trials=30,
        max_bootstrapped_demos=4,
        max_labeled_demos=16,
        seed=9,
        minibatch=True,
        minibatch_size=25,
        minibatch_full_eval_steps=10,
        program_aware_proposer=True,
        data_aware_proposer=True,
        view_data_batch_size=10,
        tip_aware_proposer=True,
        fewshot_aware_proposer=True,
        requires_permission_to_run=True,
    ):
        # Define ANSI escape codes for colors
        YELLOW = "\033[93m"
        GREEN = "\033[92m"
        BLUE = "\033[94m"
        BOLD = "\033[1m"
        ENDC = "\033[0m"  # Resets the color to default

        random.seed(seed)

        # Validate inputs
        if not trainset:
            raise ValueError("Trainset cannot be empty.")

        if not valset:
            if len(trainset) < 2:
                raise ValueError("Trainset must have at least 2 examples if no valset specified, or at least 1 example with external validation set.")
            
            valset_size = min(500, max(1, int(len(trainset) * 0.80))) # 80% of trainset, capped at 500
            cutoff = len(trainset) - valset_size
            valset = trainset[cutoff:]
            trainset = trainset[:cutoff]

        else:
            if len(valset) < 1:
                raise ValueError("Validation set must have at least 1 example if specified.")
            
        if minibatch and minibatch_size > len(valset):
            raise ValueError(f"Minibatch size cannot exceed the size of the valset.  Note that your validation set contains {len(valset)} examples.  Your train set contains {len(trainset)} examples.")
        
        if minibatch and num_trials < minibatch_full_eval_steps:
            raise ValueError(f"Number of trials (num_trials={num_trials}) must be greater than or equal to the number of minibatch full eval steps (minibatch_full_eval_steps={minibatch_full_eval_steps}).")

        estimated_prompt_model_calls = 10 + self.num_candidates * len(
                student.predictors(),
            ) + (0 if not program_aware_proposer else len(student.predictors()) + 1)  # num data summary calls + N * P + (P + 1)
        
        prompt_model_line = ""
        if not program_aware_proposer:
            prompt_model_line = f"""{YELLOW}- Prompt Model: {BLUE}{BOLD}10{ENDC}{YELLOW} data summarizer calls + {BLUE}{BOLD}{self.num_candidates}{ENDC}{YELLOW} * {BLUE}{BOLD}{len(student.predictors())}{ENDC}{YELLOW} lm calls in program = {BLUE}{BOLD}{estimated_prompt_model_calls}{ENDC}{YELLOW} prompt model calls{ENDC}"""
        else:
            prompt_model_line = f"""{YELLOW}- Prompt Model: {BLUE}{BOLD}10{ENDC}{YELLOW} data summarizer calls + {BLUE}{BOLD}{self.num_candidates}{ENDC}{YELLOW} * {BLUE}{BOLD}{len(student.predictors())}{ENDC}{YELLOW} lm calls in program + ({BLUE}{BOLD}{len(student.predictors()) + 1}{ENDC}{YELLOW}) lm calls in program aware proposer = {BLUE}{BOLD}{estimated_prompt_model_calls}{ENDC}{YELLOW} prompt model calls{ENDC}"""

        estimated_task_model_calls_wo_module_calls = 0
        task_model_line = ""
        if not minibatch:
            estimated_task_model_calls_wo_module_calls = len(trainset) * num_trials  # M * T * P
            task_model_line = f"""{YELLOW}- Task Model: {BLUE}{BOLD}{len(valset)}{ENDC}{YELLOW} examples in val set * {BLUE}{BOLD}{num_trials}{ENDC}{YELLOW} batches * {BLUE}{BOLD}# of LM calls in your program{ENDC}{YELLOW} = ({BLUE}{BOLD}{estimated_task_model_calls_wo_module_calls} * # of LM calls in your program{ENDC}{YELLOW}) task model calls{ENDC}"""
        else:
            estimated_task_model_calls_wo_module_calls = minibatch_size * num_trials + (len(trainset) * (num_trials // minibatch_full_eval_steps))  # B * T * P
            task_model_line = f"""{YELLOW}- Task Model: {BLUE}{BOLD}{minibatch_size}{ENDC}{YELLOW} examples in minibatch * {BLUE}{BOLD}{num_trials}{ENDC}{YELLOW} batches + {BLUE}{BOLD}{len(valset)}{ENDC}{YELLOW} examples in val set * {BLUE}{BOLD}{num_trials // minibatch_full_eval_steps}{ENDC}{YELLOW} full evals = {BLUE}{BOLD}{estimated_task_model_calls_wo_module_calls}{ENDC}{YELLOW} task model calls{ENDC}"""
            

        user_message = textwrap.dedent(f"""\
            {YELLOW}{BOLD}Projected Language Model (LM) Calls{ENDC}

            Please be advised that based on the parameters you have set, the maximum number of LM calls is projected as follows:

            
            {prompt_model_line}
            {task_model_line}

            {YELLOW}{BOLD}Estimated Cost Calculation:{ENDC}

            {YELLOW}Total Cost = (Number of calls to task model * (Avg Input Token Length per Call * Task Model Price per Input Token + Avg Output Token Length per Call * Task Model Price per Output Token) 
                        + (Number of calls to prompt model * (Avg Input Token Length per Call * Task Prompt Price per Input Token + Avg Output Token Length per Call * Prompt Model Price per Output Token).{ENDC}

            For a preliminary estimate of potential costs, we recommend you perform your own calculations based on the task
            and prompt models you intend to use. If the projected costs exceed your budget or expectations, you may consider:

            {YELLOW}- Reducing the number of trials (`num_trials`), the size of the valset, or the number of LM calls in your program.{ENDC}
            {YELLOW}- Using a cheaper task model to optimize the prompt.{ENDC}\n""")

        user_confirmation_message = textwrap.dedent(f"""\
            To proceed with the execution of this program, please confirm by typing {BLUE}'y'{ENDC} for yes or {BLUE}'n'{ENDC} for no.

            If you would like to bypass this confirmation step in future executions, set the {YELLOW}`requires_permission_to_run`{ENDC} flag to {YELLOW}`False`{ENDC} when calling compile.

            {YELLOW}Awaiting your input...{ENDC}
        """)

        if requires_permission_to_run: print(user_message)

        sys.stdout.flush()  # Flush the output buffer to force the message to print

        run = True
        # TODO: make sure these estimates are good for mini-batching
        if requires_permission_to_run:
            print(user_confirmation_message)
            user_input = input("Do you wish to continue? (y/n): ").strip().lower()
            if user_input != "y":
                print("Compilation aborted by the user.")
                run = False

        if run:
            # Setup random seeds
            random.seed(seed)
            np.random.seed(seed)

            # Set up program and evaluation function
            program = student.deepcopy()
            evaluate = Evaluate(
                devset=valset,
                metric=self.metric,
                num_threads=self.num_threads,
                max_errors=self.max_errors,
                display_table=False,
                display_progress=True,
            )

            # Determine the number of fewshot examples to use to generate demos for prompt
            if max_bootstrapped_demos == 0 and max_labeled_demos == 0:
                max_bootstrapped_demos_for_candidate_gen = BOOTSTRAPPED_FEWSHOT_EXAMPLES_IN_CONTEXT
                max_labeled_demos_for_candidate_gen = LABELED_FEWSHOT_EXAMPLES_IN_CONTEXT
            else:
                max_bootstrapped_demos_for_candidate_gen = max_bootstrapped_demos
                max_labeled_demos_for_candidate_gen = max_labeled_demos
            
            # Generate N few shot example sets (these will inform instruction creation, and be used as few-shot examples in our prompt)
            print("Beginning MIPROv2 optimization process...")
            print("\n==> STEP 1: BOOTSTRAP FEWSHOT EXAMPLES <==")
            if max_bootstrapped_demos > 0:
                print("These will be used for as few-shot examples candidates for our program and for creating instructions.\n")
            else:
                print("These will be used for informing instruction proposal.\n")
            print(f"Bootstrapping N={self.num_candidates} sets of demonstrations...")
            try:
                demo_candidates = create_n_fewshot_demo_sets(
                    student=program,
                    num_candidate_sets=self.num_candidates,
                    trainset=trainset,
                    max_labeled_demos=max_labeled_demos_for_candidate_gen,
                    max_bootstrapped_demos=max_bootstrapped_demos_for_candidate_gen,
                    metric=self.metric,
                    max_errors=self.max_errors,
                    teacher_settings=self.teacher_settings,
                    seed=seed,
                    metric_threshold=self.metric_threshold,
                )
            except Exception as e:
                print(f"Error generating fewshot examples: {e}")
                print("Running without fewshot examples.")
                demo_candidates = None
            
            # Generate N candidate instructions

            # Setup our proposer 
            print("\n==> STEP 2: PROPOSE INSTRUCTION CANDIDATES <==")
            print("In this step, by default we will use the few-shot examples from the previous step, a generated dataset summary, a summary of the program code, and a randomly selected prompting tip to propose instructions.")

            proposer = GroundedProposer(
                program=program,
                trainset=trainset,
                prompt_model=self.prompt_model,
                # program_code_string=self.program_code_string,
                view_data_batch_size=view_data_batch_size,
                program_aware=program_aware_proposer,
                use_dataset_summary=data_aware_proposer,
                use_task_demos=fewshot_aware_proposer,
                use_tip=tip_aware_proposer,
                set_tip_randomly=tip_aware_proposer,
                use_instruct_history=False,
                set_history_randomly=False,
                verbose = self.verbose,
            )
            
            print("\nProposing instructions...\n")
            instruction_candidates = proposer.propose_instructions_for_program(
                trainset=trainset,
                program=program,
                demo_candidates=demo_candidates,
                N=self.num_candidates,
                T=self.init_temperature,
                trial_logs={},
            )
            for i, pred in enumerate(program.predictors()):
                print(f"Proposed Instructions for Predictor {i}:\n")
                instruction_candidates[i][0] = get_signature(pred).instructions
                for j, instruction in enumerate(instruction_candidates[i]):
                    print(f"{j}: {instruction}\n")
                print("\n")

            # If we're doing zero-shot, reset demo_candidates to none now that we've used them for instruction proposal
            if max_bootstrapped_demos == 0 and max_labeled_demos == 0:
                demo_candidates = None

            # Initialize variables to track during the optimization process
            best_scoring_trial = 0
            trial_logs = {}
            total_eval_calls = 0
            param_score_dict = defaultdict(list) # Dictionaries of paramater combinations we've tried, and their associated scores
            fully_evaled_param_combos = {} # List of the parameter combinations we've done full evals of

            # Evaluate the default program
            print("Evaluating the default program...\n")
            default_score = eval_candidate_program(len(valset), valset, program, evaluate)
            print(f"Default program score: {default_score}\n")

            best_score = default_score
            best_program = program.deepcopy()


            # Define our trial objective
            def create_objective(
                baseline_program,
                instruction_candidates,
                demo_candidates,
                evaluate,
                valset,
            ):
                def objective(trial):
                    nonlocal best_program, best_score, best_scoring_trial, trial_logs, total_eval_calls  # Allow access to the outer variables
                    
                    # Kick off trial
                    if minibatch:
                        print(f"== Minibatch Trial {trial.number+1} / {num_trials} ==")
                    else:
                        print(f"===== Trial {trial.number+1} / {num_trials} =====")
                    trial_logs[trial.number+1] = {}

                    # Create a new candidate program
                    candidate_program = baseline_program.deepcopy()

                    # Choose set of instructions & demos to use for each predictor
                    chosen_params = []
                    for i, p_new in enumerate(candidate_program.predictors()):

                        # Get instruction candidates / demos for our given predictor
                        p_instruction_candidates = instruction_candidates[i]
                        if demo_candidates:
                            p_demo_candidates = demo_candidates[i]

                        # Suggest the index of the instruction / demo candidate to use in our trial
                        instruction_idx = trial.suggest_categorical(
                            f"{i}_predictor_instruction",
                            range(len(p_instruction_candidates)),
                        )
                        # chosen_params.append(instruction_idx)
                        chosen_params.append(f"Predictor {i+1}: Instruction {instruction_idx}")
                        if demo_candidates:
                            demos_idx = trial.suggest_categorical(
                                f"{i}_predictor_demos", range(len(p_demo_candidates)),
                            )
                            chosen_params.append(f"Predictor {i+1}: Few-Shot Set {demos_idx}")

                        # Log the selected instruction / demo candidate
                        trial_logs[trial.number+1][
                            f"{i}_predictor_instruction"
                        ] = instruction_idx
                        if demo_candidates:
                            trial_logs[trial.number+1][f"{i}_predictor_demos"] = demos_idx

                        dspy.logger.debug(f"instruction_idx {instruction_idx}")
                        if demo_candidates:
                            dspy.logger.debug(f"demos_idx {demos_idx}")

                        # Set the instruction
                        selected_instruction = p_instruction_candidates[instruction_idx]
                        updated_signature = get_signature(p_new).with_instructions(
                            selected_instruction,
                        )
                        set_signature(p_new, updated_signature)

                        # Set the demos
                        if demo_candidates:
                            p_new.demos = p_demo_candidates[demos_idx]

                    # Log assembled program
                    if self.verbose: print("Evaluating the following candidate program...\n")
                    if self.verbose: print_full_program(candidate_program)

                    # Save the candidate program
                    trial_logs[trial.number+1]["program_path"] = save_candidate_program(
                        candidate_program, self.log_dir, trial.number+1,
                    )

                    trial_logs[trial.number+1]["num_eval_calls"] = 0

                    # Evaluate the candidate program with relevant batch size
                    batch_size = minibatch_size if minibatch else len(valset)

                    score = eval_candidate_program(
                        batch_size, valset, candidate_program, 
                        evaluate,
                    )

                    # Print out a full trace of the program in use
                    if self.verbose:
                        print("Full trace of prompts in use on an example...")
                        get_task_model_history_for_full_example(
                            candidate_program, self.task_model, valset, evaluate,
                        )

                    # Log relevant information
                    categorical_key = ",".join(map(str, chosen_params))
                    param_score_dict[categorical_key].append(
                        (score, candidate_program),
                    )
                    trial_logs[trial.number+1]["num_eval_calls"] = batch_size
                    trial_logs[trial.number+1]["full_eval"] = batch_size >= len(valset)
                    trial_logs[trial.number+1]["score"] = score
                    trial_logs[trial.number+1]["pruned"] = False
                    total_eval_calls += trial_logs[trial.number+1]["num_eval_calls"]
                    trial_logs[trial.number+1]["total_eval_calls_so_far"] = total_eval_calls
                    trial_logs[trial.number+1]["program"] = candidate_program.deepcopy()

                    # If this score was from a full evaluation, update the best program if the new score is better
                    best_score_updated = False
                    if score > best_score and trial_logs[trial.number+1]["full_eval"] and not minibatch:
                        best_score = score
                        best_scoring_trial = trial.number+1
                        best_program = candidate_program.deepcopy()
                        best_score_updated = True
                    
                    if minibatch:
                        print(f"Score: {score} on minibatch of size {batch_size} with parameters {chosen_params}.\n\n")
                    else:
                        print(f"Score: {score} with parameters {chosen_params}.")
                        if best_score_updated:
                            print(f"{GREEN}New best score updated!{ENDC} Score: {best_score} on trial {best_scoring_trial}.\n\n")
                        else:
                            print(f"Best score so far: {best_score} on trial {best_scoring_trial}.\n\n")
                        

                    # If we're doing minibatching, check to see if it's time to do a full eval
                    if minibatch and (((trial.number+1) % minibatch_full_eval_steps == 0) or (trial.number+1 == num_trials)):
                        print(f"===== Full Eval {len(fully_evaled_param_combos)+1} =====")
                        
                        # Save old information as the minibatch version
                        trial_logs[trial.number+1]["mb_score"] = score
                        trial_logs[trial.number+1]["mb_program_path"] = trial_logs[trial.number+1]["program_path"]

                        # Identify our best program (based on mean of scores so far, and do a full eval on it)
                        highest_mean_program, mean, combo_key = get_program_with_highest_avg_score(param_score_dict, fully_evaled_param_combos)
                        
                        if trial.number+1 // minibatch_full_eval_steps > 0:
                            print(f"Doing full eval on next top averaging program (Avg Score: {mean}) so far from mini-batch trials...")
                        else:
                            print(f"Doing full eval on top averaging program (Avg Score: {mean}) so far from mini-batch trials...")
                        full_val_score = eval_candidate_program(
                            len(valset), valset, highest_mean_program, evaluate,
                        )

                        # Log relevant information
                        fully_evaled_param_combos[combo_key] = {"program":highest_mean_program, "score": full_val_score}
                        total_eval_calls += len(valset)
                        trial_logs[trial.number+1]["total_eval_calls_so_far"] = total_eval_calls
                        trial_logs[trial.number+1]["full_eval"] = True
                        trial_logs[trial.number+1]["program_path"] = save_candidate_program(
                            program=highest_mean_program, log_dir=self.log_dir, trial_num=trial.number+1, note="full_eval",
                        )
                        trial_logs[trial.number+1]["score"] = full_val_score
                        
                        if full_val_score > best_score:
                            print(f"{GREEN}Best full eval score so far!{ENDC} Score: {full_val_score}")
                            best_score = full_val_score
                            best_scoring_trial = trial.number+1
                            best_program = highest_mean_program.deepcopy()
                            best_score_updated = True
                        else:
                            print(f"Full eval score: {full_val_score}")
                            print(f"Best full eval score so far: {best_score}")
                        print("=======================\n\n")
                    
                    return score

                return objective

            # Run the trial
            optuna.logging.set_verbosity(optuna.logging.WARNING)
            print("==> STEP 3: FINDING OPTIMAL PROMPT PARAMETERS <==")
            print("In this step, we will evaluate the program over a series of trials with different combinations of instructions and few-shot examples to find the optimal combination. Bayesian Optimization will be used for this search process.\n")
            objective_function = create_objective(
                program, instruction_candidates, demo_candidates, evaluate, valset,
            )
            sampler = optuna.samplers.TPESampler(seed=seed, multivariate=True)
            study = optuna.create_study(direction="maximize", sampler=sampler)
            _ = study.optimize(objective_function, n_trials=num_trials)

            if best_program is not None and self.track_stats:
                best_program.trial_logs = trial_logs
                best_program.score = best_score
                best_program.prompt_model_total_calls = self.prompt_model_total_calls
                best_program.total_calls = self.total_calls

            return best_program

        return student



import logging
import random
from typing import List, Optional

from pydantic import BaseModel

import dspy


class DescriptionSignature(dspy.Signature):
    field_name = dspy.InputField(desc="name of a field")
    example = dspy.InputField(desc="an example value for the field")
    description = dspy.OutputField(desc="a short text only description of what the field contains")


class SyntheticDataGenerator:
    def __init__(self, schema_class: Optional[BaseModel] = None, examples: Optional[List[dspy.Example]] = None):
        self.schema_class = schema_class
        self.examples = examples

    def generate(self, sample_size: int) -> List[dspy.Example]:
        """Generate synthetic examples.

        Args:
            sample_size (int): number of examples to generate
        Raises:
            ValueError: either a schema_class or examples should be provided
        Returns:
            List[dspy.Example]: list of synthetic examples generated
        """
        if not self.schema_class and not self.examples:
            raise ValueError("Either a schema_class or examples must be provided.")
        if self.examples and len(self.examples) >= sample_size:
            logging.info("No additional data generation needed.")
            return self.examples[:sample_size]

        additional_samples_needed = sample_size - (len(self.examples) if self.examples else 0)
        generated_examples = self._generate_additional_examples(additional_samples_needed)

        return self.examples + generated_examples if self.examples else generated_examples

    def _define_or_infer_fields(self):
        """Define fields to generate if a schema class is provided.
        Infer fields to generate if an inital sample of examples is provided.

        Returns:
            dict: dictionary of fields to generate
        """  # noqa: D205
        if self.schema_class:
            data_schema = self.schema_class.model_json_schema()
            properties = data_schema['properties']
        elif self.examples:
            inferred_schema = self.examples[0].__dict__['_store']
            descriptor = dspy.Predict(DescriptionSignature)
            properties = {field: {'description': str((descriptor(field_name=field, example=str(inferred_schema[field]))).description)}
                          for field in inferred_schema.keys()}
        else:
            properties = {}
        return properties

    def _generate_additional_examples(self, additional_samples_needed: int) -> List[dspy.Example]:
        """Generate additional examples if needed.

        Args:
            additional_samples_needed (int): the difference between the desired
            number of examples and the current number of examples
        Returns:
            List[dspy.Example]: list of synthetic examples
        """
        properties = self._define_or_infer_fields()
        class_name = f"{self.schema_class.__name__ if self.schema_class else 'Inferred'}Signature"
        fields = self._prepare_fields(properties)

        signature_class = type(class_name, (dspy.Signature,), fields)
        generator = dspy.Predict(signature_class, n=additional_samples_needed)
        response = generator(sindex=str(random.randint(1, additional_samples_needed)))

        return [dspy.Example({field_name: getattr(completion, field_name) for field_name in properties.keys()})
                for completion in response.completions]

    def _prepare_fields(self, properties) -> dict:
        """Prepare fields to generate in an appropriate format."""
        return {
            '__doc__': f"Generates the following outputs: {{{', '.join(properties.keys())}}}.",
            'sindex': dspy.InputField(desc="a random string"),
            **{field_name: dspy.OutputField(desc=properties[field_name].get('description', 'No description'))
               for field_name in properties.keys()},
        }

# # Usage example
# # Generating synthetic data via a pydantic model
# generator = SyntheticDataGenerator(schema_class=SyntheticFacts)
# examples = generator.generate(sample_size=6)

# # Generating synthetic data via existing examples
# generator = SyntheticDataGenerator(examples=existing_examples)
# examples = generator.generate(sample_size=5)



import os
import ujson
import functools
from pathlib import Path

try:
    import warnings
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        if "LITELLM_LOCAL_MODEL_COST_MAP" not in os.environ:
             os.environ["LITELLM_LOCAL_MODEL_COST_MAP"] = "True"
        import litellm  
        litellm.telemetry = False

    from litellm.caching import Cache
    disk_cache_dir = os.environ.get('DSPY_CACHEDIR') or os.path.join(Path.home(), '.dspy_cache')
    litellm.cache = Cache(disk_cache_dir=disk_cache_dir, type="disk")

except ImportError:
    class LitellmPlaceholder:
        def __getattr__(self, _): raise ImportError("The LiteLLM package is not installed. Run `pip install litellm`.")

    litellm = LitellmPlaceholder()

class LM:
    def __init__(self, model, model_type='chat', temperature=0.0, max_tokens=1000, cache=True, **kwargs):
        self.model = model
        self.model_type = model_type
        self.cache = cache
        self.kwargs = dict(temperature=temperature, max_tokens=max_tokens, **kwargs)
        self.history = []

        if "o1-" in model:
            assert max_tokens >= 5000 and temperature == 1.0, \
                "OpenAI's o1-* models require passing temperature=1.0 and max_tokens >= 5000 to `dspy.LM(...)`"
                
    
    def __call__(self, prompt=None, messages=None, **kwargs):
        # Build the request.
        cache = kwargs.pop("cache", self.cache)
        messages = messages or [{"role": "user", "content": prompt}]
        kwargs = {**self.kwargs, **kwargs}

        # Make the request and handle LRU & disk caching.
        if self.model_type == "chat": completion = cached_litellm_completion if cache else litellm_completion
        else: completion = cached_litellm_text_completion if cache else litellm_text_completion

        response = completion(ujson.dumps(dict(model=self.model, messages=messages, **kwargs)))
        outputs = [c.message.content if hasattr(c, "message") else c["text"] for c in response["choices"]]

        # Logging, with removed api key & where `cost` is None on cache hit.
        kwargs = {k: v for k, v in kwargs.items() if not k.startswith("api_")}
        entry = dict(prompt=prompt, messages=messages, kwargs=kwargs, response=response)
        entry = dict(**entry, outputs=outputs, usage=dict(response["usage"]))
        entry = dict(**entry, cost=response.get("_hidden_params", {}).get("response_cost"))
        self.history.append(entry)

        return outputs
    
    def inspect_history(self, n: int = 1):
        _inspect_history(self, n)


@functools.lru_cache(maxsize=None)
def cached_litellm_completion(request):
    return litellm_completion(request, cache={"no-cache": False, "no-store": False})

def litellm_completion(request, cache={"no-cache": True, "no-store": True}):
    kwargs = ujson.loads(request)
    return litellm.completion(cache=cache, **kwargs)

@functools.lru_cache(maxsize=None)
def cached_litellm_text_completion(request):
    return litellm_text_completion(request, cache={"no-cache": False, "no-store": False})

def litellm_text_completion(request, cache={"no-cache": True, "no-store": True}):
    kwargs = ujson.loads(request)

    # Extract the provider and model from the model string.
    model = kwargs.pop("model").split("/", 1)
    provider, model = model[0] if len(model) > 1 else "openai", model[-1]

    # Use the API key and base from the kwargs, or from the environment.
    api_key = kwargs.pop("api_key", None) or os.getenv(f"{provider}_API_KEY")
    api_base = kwargs.pop("api_base", None) or os.getenv(f"{provider}_API_BASE")

    # Build the prompt from the messages.
    prompt = '\n\n'.join([x['content'] for x in kwargs.pop("messages")] + ['BEGIN RESPONSE:'])

    return litellm.text_completion(cache=cache, model=f'text-completion-openai/{model}', api_key=api_key,
                                   api_base=api_base, prompt=prompt, **kwargs)


def _green(text: str, end: str = "\n"):
    return "\x1b[32m" + str(text).lstrip() + "\x1b[0m" + end

def _red(text: str, end: str = "\n"):
    return "\x1b[31m" + str(text) + "\x1b[0m" + end

def _inspect_history(lm, n: int = 1):
    """Prints the last n prompts and their completions."""

    for item in lm.history[-n:]:
        messages = item["messages"] or [{"role": "user", "content": item['prompt']}]
        outputs = item["outputs"]

        print("\n\n\n")
        for msg in messages:
            print(_red(f"{msg['role'].capitalize()} message:"))
            print(msg['content'].strip())
            print("\n")

        print(_red("Response:"))
        print(_green(outputs[0].strip()))

        if len(outputs) > 1:
            choices_text = f" \t (and {len(outputs)-1} other completions)"
            print(_red(choices_text, end=""))
        
    print("\n\n\n")



import threading
from contextlib import contextmanager

from dsp.utils.utils import dotdict


class Settings:
    """DSP configuration settings."""

    _instance = None

    def __new__(cls):
        """
        Singleton Pattern. See https://python-patterns.guide/gang-of-four/singleton/
        """

        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.lock = threading.Lock()
            cls._instance.main_tid = threading.get_ident()
            cls._instance.main_stack = []
            cls._instance.stack_by_thread = {}
            cls._instance.stack_by_thread[threading.get_ident()] = cls._instance.main_stack

            #  TODO: remove first-class support for re-ranker and potentially combine with RM to form a pipeline of sorts
            #  eg: RetrieveThenRerankPipeline(RetrievalModel, Reranker)
            #  downstream operations like dsp.retrieve would use configs from the defined pipeline.
            config = dotdict(
                lm=None,
                adapter=None,
                rm=None,
                branch_idx=0,
                reranker=None,
                compiled_lm=None,
                force_reuse_cached_compilation=False,
                compiling=False,  # TODO: can probably be removed
                skip_logprobs=False,
                trace=[],
                release=0,
                bypass_assert=False,
                bypass_suggest=False,
                assert_failures=0,
                suggest_failures=0,
                langchain_history=[],
                experimental=False,
                backoff_time = 10
            )
            cls._instance.__append(config)

        return cls._instance

    @property
    def config(self):
        thread_id = threading.get_ident()
        if thread_id not in self.stack_by_thread:
            self.stack_by_thread[thread_id] = [self.main_stack[-1].copy()]
        return self.stack_by_thread[thread_id][-1]

    def __getattr__(self, name):
        if hasattr(self.config, name):
            return getattr(self.config, name)

        if name in self.config:
            return self.config[name]

        super().__getattr__(name)

    def __append(self, config):
        thread_id = threading.get_ident()
        if thread_id not in self.stack_by_thread:
            self.stack_by_thread[thread_id] = [self.main_stack[-1].copy()]
        self.stack_by_thread[thread_id].append(config)

    def __pop(self):
        thread_id = threading.get_ident()
        if thread_id in self.stack_by_thread:
            self.stack_by_thread[thread_id].pop()

    def configure(self, inherit_config: bool = True, **kwargs):
        """Set configuration settings.

        Args:
            inherit_config (bool, optional): Set configurations for the given, and use existing configurations for the rest. Defaults to True.
        """
        if inherit_config:
            config = {**self.config, **kwargs}
        else:
            config = {**kwargs}

        self.__append(config)

    @contextmanager
    def context(self, inherit_config=True, **kwargs):
        self.configure(inherit_config=inherit_config, **kwargs)

        try:
            yield
        finally:
            self.__pop()

    def __repr__(self) -> str:
        return repr(self.config)


settings = Settings()



import dsp
import dspy
from dspy.signatures.signature import ensure_signature

from ..primitives.program import Module
from .predict import Predict

# TODO: Simplify a lot.
# TODO: Divide Action and Action Input like langchain does for ReAct.

# TODO: There's a lot of value in having a stopping condition in the LM calls at `\n\nObservation:`

# TODO [NEW]: When max_iters is about to be reached, reduce the set of available actions to only the Finish action.


class ReAct(Module):
    def __init__(self, signature, max_iters=5, num_results=3, tools=None):
        super().__init__()
        self.signature = signature = ensure_signature(signature)
        self.max_iters = max_iters

        self.tools = tools or [dspy.Retrieve(k=num_results)]
        self.tools = {tool.name: tool for tool in self.tools}

        self.input_fields = self.signature.input_fields
        self.output_fields = self.signature.output_fields

        assert len(self.output_fields) == 1, "ReAct only supports one output field."

        inputs_ = ", ".join([f"`{k}`" for k in self.input_fields.keys()])
        outputs_ = ", ".join([f"`{k}`" for k in self.output_fields.keys()])

        instr = []
        
        if self.signature.instructions is not None:
            instr.append(f"{self.signature.instructions}\n")
        
        instr.extend([
            f"You will be given {inputs_} and you will respond with {outputs_}.\n",
            "To do this, you will interleave Thought, Action, and Observation steps.\n",
            "Thought can reason about the current situation, and Action can be the following types:\n",
        ])

        self.tools["Finish"] = dspy.Example(
            name="Finish",
            input_variable=outputs_.strip("`"),
            desc=f"returns the final {outputs_} and finishes the task",
        )

        for idx, tool in enumerate(self.tools):
            tool = self.tools[tool]
            instr.append(
                f"({idx+1}) {tool.name}[{tool.input_variable}], which {tool.desc}",
            )

        instr = "\n".join(instr)
        self.react = [
            Predict(dspy.Signature(self._generate_signature(i), instr))
            for i in range(1, max_iters + 1)
        ]

    def _generate_signature(self, iters):
        signature_dict = {}
        for key, val in self.input_fields.items():
            signature_dict[key] = val

        for j in range(1, iters + 1):
            IOField = dspy.OutputField if j == iters else dspy.InputField

            signature_dict[f"Thought_{j}"] = IOField(
                prefix=f"Thought {j}:",
                desc="next steps to take based on last observation",
            )

            tool_list = " or ".join(
                [
                    f"{tool.name}[{tool.input_variable}]"
                    for tool in self.tools.values()
                    if tool.name != "Finish"
                ],
            )
            signature_dict[f"Action_{j}"] = IOField(
                prefix=f"Action {j}:",
                desc=f"always either {tool_list} or, when done, Finish[<answer>], where <answer> is the answer to the question itself.",
            )

            if j < iters:
                signature_dict[f"Observation_{j}"] = IOField(
                    prefix=f"Observation {j}:",
                    desc="observations based on action",
                    format=dsp.passages2text,
                )

        return signature_dict

    def act(self, output, hop):
        try:
            action = output[f"Action_{hop+1}"]
            action_name, action_val = action.strip().split("\n")[0].split("[", 1)
            action_val = action_val.rsplit("]", 1)[0]

            if action_name == "Finish":
                return action_val

            result = self.tools[action_name](action_val)  #result must be a str, list, or tuple
            # Handle the case where 'passages' attribute is missing
            output[f"Observation_{hop+1}"] = getattr(result, "passages", result)

        except Exception:
            output[f"Observation_{hop+1}"] = (
                "Failed to parse action. Bad formatting or incorrect action name."
            )
            # raise e

    def forward(self, **kwargs):
        args = {key: kwargs[key] for key in self.input_fields.keys() if key in kwargs}

        for hop in range(self.max_iters):
            # with dspy.settings.context(show_guidelines=(i <= 2)):
            output = self.react[hop](**args)
            output[f'Action_{hop + 1}'] = output[f'Action_{hop + 1}'].split('\n')[0]

            if action_val := self.act(output, hop):
                break
            args.update(output)

        observations = [args[key] for key in args if key.startswith("Observation")]
        
        # assumes only 1 output field for now - TODO: handling for multiple output fields
        return dspy.Prediction(observations=observations, **{list(self.output_fields.keys())[0]: action_val or ""})



from abc import ABC, abstractmethod


class LM(ABC):
    """Abstract class for language models."""

    def __init__(self, model, tracker=None):
        self.kwargs = {
            "model": model,
            "temperature": 0.0,
            "max_tokens": 150,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "n": 1,
        }
        self.provider = "default"
        self.tracker = tracker

        self.history = []

    @abstractmethod
    def basic_request(self, prompt, **kwargs):
        pass

    def request(self, prompt, **kwargs):
        return self.basic_request(prompt, **kwargs)

    def print_green(self, text: str, end: str = "\n"):
        import dspy

        if dspy.settings.experimental:
            return "\n\n" + "\x1b[32m" + str(text).lstrip() + "\x1b[0m" + end
        else:
            return "\x1b[32m" + str(text) + "\x1b[0m" + end

    def print_red(self, text: str, end: str = "\n"):
        return "\x1b[31m" + str(text) + "\x1b[0m" + end

    def inspect_history(self, n: int = 1, skip: int = 0, color_format: bool = True):
        """Prints the last n prompts and their completions.

        TODO: print the valid choice that contains filled output field instead of the first.
        """
        provider: str = self.provider

        last_prompt = None
        printed = []
        n = n + skip

        for x in reversed(self.history[-100:]):
            prompt = x["prompt"]

            if prompt != last_prompt:
                if provider in (
                    "clarifai",
                    "cloudflare",
                    "google",
                    "groq",
                    "Bedrock",
                    "Sagemaker",
                    "premai",
                    "tensorrt_llm",
                ):
                    printed.append((prompt, x["response"]))
                elif provider == "anthropic":
                    blocks = [
                        {"text": block.text}
                        for block in x["response"].content
                        if block.type == "text"
                    ]
                    printed.append((prompt, blocks))
                elif provider == "cohere":
                    printed.append((prompt, x["response"].text))
                elif provider == "mistral":
                    printed.append((prompt, x["response"].choices))
                elif provider == "ibm":
                    printed.append((prompt, x))
                elif provider == "you.com":
                    printed.append((prompt, x["response"]["answer"]))
                else:
                    printed.append((prompt, x["response"]["choices"]))

            last_prompt = prompt

            if len(printed) >= n:
                break

        printing_value = ""
        for idx, (prompt, choices) in enumerate(reversed(printed)):
            # skip the first `skip` prompts
            if (n - idx - 1) < skip:
                continue
            printing_value += "\n\n\n"
            printing_value += prompt

            text = ""
            if provider in (
                "cohere",
                "Bedrock",
                "Sagemaker",
                "clarifai",
                "claude",
                "ibm",
                "premai",
                "you.com",
                "tensorrt_llm",
            ):
                text = choices
            elif provider == "openai" or provider == "ollama" or provider == "llama":
                text = " " + self._get_choice_text(choices[0]).strip()
            elif provider == "groq":
                text = " " + choices
            elif provider == "google":
                text = choices[0].parts[0].text
            elif provider == "mistral":
                text = choices[0].message.content
            elif provider == "cloudflare":
                text = choices[0]
            else:
                text = choices[0]["text"]
            printing_value += self.print_green(text, end="") if color_format else text

            if len(choices) > 1 and isinstance(choices, list):
                choices_text = f" \t (and {len(choices)-1} other completions)"
                printing_value += self.print_red(
                   choices_text, end="",
                ) if color_format else choices_text

            printing_value += "\n\n\n"

        print(printing_value)
        return printing_value

    @abstractmethod
    def __call__(self, prompt, only_completed=True, return_sorted=False, **kwargs):
        pass

    def tracker_call(self, tracker, prompt=None, output=None, name=None, **kwargs):
        from dsp.trackers.base import BaseTracker
        assert issubclass(tracker.__class__, BaseTracker), "tracker must be a subclass of BaseTracker"
        assert self.history, "tracker.call() requires a previous request"

        last_req = self.history[-1]
        if not prompt:
            prompt = last_req.get('prompt', None)
        if not output:
            output = last_req.get('response', None)
        kwargs = {**self.kwargs, **kwargs}
        name = name if name else self.__class__.__name__
        tracker.call(i=prompt, o=output, name=name, **kwargs)

    def copy(self, **kwargs):
        """Returns a copy of the language model with the same parameters."""
        kwargs = {**self.kwargs, **kwargs}
        model = kwargs.pop("model")

        return self.__class__(model=model, **kwargs)



""" 
TODO: If we want to have Prediction::{**keys, completions, box} where box.{key} will behave as a value but also include
the completions internally.

The main thing left is to determine the semantic (and then implement them) for applying operations on the object.

If we have a string (query) and completions (5 queries), and we modify the string, what happens to the completions?

- Option 1: We modify the string and the (other) completions are not affected.
- Option 2: We modify the string and the (other) completions too.
- Option 3: We modify the string and the (other) completions are deleted.

Option 2 seems most reasonable, but it depends on what metadata the box is going to store.

It seems that a box fundamentally has two functions then:
- Store a value and its "alternatives" (and hence allow transparent application over operations on value/alts)
    - But not all operations can/should be applied on all as a map.
    - I guess mainly "give me a string or list or dict or tuple or int or float" has to commit to a value.
    - There also needs to be a .item().
- Potentially track the "source" of the value (i.e., the predictor that generated it, and its inputs)
- Give the value (eventually) to something that will consume the main value (implicitly or explicitly) or all/some of its alternatives explicitly.

It might be wise to make this responsible for a smaller scope for now:

- Just one string (and its alternatives).
- No source tracking.
- Allow operations on the string to map over the alternatives.
- Seamless extraction at code boundaries.
    - Basically, code will either treat this as string implicitly
        (and hence only know about the one value, and on best effort basis we update the alternatives)
    - Or code will explicitly work with the string or explicitly work with the full set of alternatives.

- By default, all programs (and their sub-programs) will be running inside a context in which preserve_boxes=True.
- But outside the program, once we see that none of the parent contexts have preserve_boxes=True, we can automatically
    unpack all boxes before returning to user.

Okay, so we'll have predictors return a `pred` in which `pred.query` is a box.

You'd usually do one of:

### Things that just give you one string
    1- Print `pred.query` or save it in a dict somewhere or a file somewhere.
    2- Call `pred.query.item()` to get the string explicitly.
    3- Modifications in freeform Python.
        - Modify it by calling `pred.query = 'new query'` altogether.
        - Modify it by doing `pred.query += 'new query'` or templating `f'{pred.query} new query'`.
        - Other modifications are not allowed on strings (e.g., `pred.query[0] = 'a'` or `pred.query[0] += 'a'`).
        - Cast to boolean after a comparison: `if pred.query == 'something': ...`
            - Pytorch would say RuntimeError: Boolean value of Tensor with more than one value is ambiguous
            - But we can keep the primary value and use that in the boolean.
            - So technically, comparison can stick around, giving you multiple internal bools.

Overall, I think it's coherent semantics, for the time being, to say that any of the above will just give you a string back and lose all tracking.


### Things that give you a list of strings
    1- Explicitly asking for the candidates/completions.
    2- Then you could filter or map that list arbitrarily.

In this case, it's just that Box will serve as syntactic sugar. If you don't want to think about `n` at all, you can
pretend you have a string. If you do anything arbitrary on it, it indeed becomes a string.
If you later decide to treat it as a list, it's easy to do so without losing that info when you say `pred.query`.
    
### Things that are more interesting

A) You can now pass pred.query to a DSPy predictor (or searcher, etc) and it can either naively work with the string,
like pass it to a template, or it can explicitly ask for the list of candidates and do something with that.

This will need a lot more string-specific operations though:
- endswith, startswith, contains, split, strip, lower, upper, etc.
- when doing ' '.join() must do map(str, values_to_join). No implicit __str__ conversion!
- We can probably automate this by having a general fallback? That either returns one value or maps that over all of them.

B) When you say dspy.assert pred.sentence1.endswith('blue'), it will actually check all the alternatives and locally filter them if possible.
It may keep the bad ones somewhere just in case too.

We could make this a little more explicit like dspy.assert(pred.sentence1, lambda x: x.endswith('blue'))

C) When program_temperature is high, we can actually have some more interesting logic here. When you try to do things that are "selective",
maybe we'll randomly give you one of the strings (that remain valid in the box, based on assertions).

This could lead to interesting efficiency, because we can basically rerun the program, it'll still generate n=10 candidates,
but each time it'll use a different one. So when branch_index changes, you get a new candidate each time, but it should be consistent in the same box.
I guess this could be done by shuffling the same N=10 things. So basically, there's a user N and there's a system-level M.

We can sometimes optimize things by doing M=5. So we'll generate Nx5 candidates in one or more calls (depending on value of Nx5).
Then depending on the branch_idx, we'll return a fixed set of N candidates. But we have the rest.

"""


class BoxType(type):
    # List of operations to override
    ops = [
        # Arithmetic operations
        'add', 'sub', 'mul', 'truediv', 'floordiv', 'mod', 'pow', 
        'lshift', 'rshift', 'and', 'or', 'xor',
        # 'r'-prefixed versions of arithmetic operations
        'radd', 'rsub', 'rmul', 'rtruediv', 'rfloordiv', 'rmod', 
        'rpow', 'rlshift', 'rrshift', 'rand', 'ror', 'rxor',
        # Sequence operations
        'getitem', 'setitem', 'delitem', 'contains',
        # Unary and other operations
        'neg', 'pos', 'abs', 'invert', 'round', 'len', 
        'getitem', 'setitem', 'delitem', 'contains', 'iter',
        # Mappings operations (for dicts)
        'get', 'keys', 'values', 'items',
        # Comparison
        'eq', 'ne', 'lt', 'le', 'gt', 'ge',
    ]

    def __init__(cls, name, bases, attrs):
        def create_method(op):
            def method(self, other=None):
                if op in ['len', 'keys', 'values', 'items']:
                    return getattr(self._value, op)()
                elif isinstance(other, Box):
                    return Box(getattr(self._value, f'__{op}__')(other._value))
                elif other is not None:
                    return Box(getattr(self._value, f'__{op}__')(other))
                else:
                    return NotImplemented
            return method

        for op in BoxType.ops:
            setattr(cls, f'__{op}__', create_method(op))

        super().__init__(name, bases, attrs)


class Box(metaclass=BoxType):
    def __init__(self, value, source=False):
        self._value = value
        self._source = source

    def __repr__(self):
        return repr(self._value)

    def __str__(self):
        return str(self._value)
    
    def __bool__(self):
        return bool(self._value)
    
    # if method is missing just call it on the _value
    def __getattr__(self, name):
        return Box(getattr(self._value, name))

    # # Unlike the others, this one collapses to a bool directly
    # def __eq__(self, other):
    #     if isinstance(other, Box):
    #         return self._value == other._value
    #     else:
    #         return self._value == other

    # def __ne__(self, other):
    #     return not self.__eq__(other)



from typing import Any, Callable

import dsp
from dsp.adapters import Template
from dsp.primitives.demonstrate import Example
from dsp.utils import zipstar


class Completions:
    """A state object that holds the valid LM completions for a given Template."""

    def __init__(self, completions: list[Example], template: Template):
        self.data = completions
        self.template = template

    def __iter__(self):
        return self.data.__iter__()

    def __getitem__(self, item):
        return self.data[item]

    def __len__(self):
        return len(self.data)

    def unpack(self, key=None):
        if key:
            return [getattr(c, key) for c in self.data]

        return zipstar(self.data)

    def __getattr__(self, name):
        assert len(self.data) == 1

        completion = self.data[0]

        if name in completion.keys():
            return getattr(completion, name)

        if name.endswith("s") and name[:-1] in completion.keys():
            pass

        assert False, name


def generate(template: Template, **kwargs) -> Callable:
    """Returns a callable function that generates completions for a given example using the provided template."""
    if hasattr(dsp.settings, "inspect"):
        inspector = dsp.settings.inspect
        _generate = inspector.inspect_func(dsp.predict._generate)
        return _generate(template, **kwargs)
    else:
        return dsp.predict._generate(template, **kwargs)


def _generate(template: Template, **kwargs) -> Callable:
    """Returns a callable function that generates completions for a given example using the provided template."""
    if not dsp.settings.lm:
        raise AssertionError("No LM is loaded.")

    generator = dsp.settings.lm

    def do_generate(example: Example, stage: str, max_depth: int = 2, original_example=None):
        if not dsp.settings.lm:
            raise AssertionError("No LM is loaded.")
        original_example = original_example or example
        assert stage is not None

        # Look up the appropriate fields in each demonstration.
        example = example.demos_at(lambda d: d[stage])

        # Generate and extract the fields.
        prompt = template(example)
        completions: list[dict[str, Any]] = generator(prompt, **kwargs)
        completions: list[Example] = [template.extract(example, p) for p in completions]

        # Find the completions that are most complete.
        field_names: list[str] = [field.input_variable for field in template.fields]

        last_field_idx = 0
        for field_idx, key in enumerate(field_names):
            completions_ = [c for c in completions if key in c.keys() and c[key] is not None]

            # Filter out completions that are missing fields that are present in at least one completion.
            if len(completions_):
                completions = completions_
                last_field_idx = field_idx + 1

        # If none of the completions is completed (i.e., none has the final field set).
        if last_field_idx < len(field_names):
            # Pick the first completion that has gone farthest.
            completion = completions[0]
            completion[field_names[last_field_idx]] = ""

            # Recurse with greedy decoding and a shorter length.
            max_tokens = kwargs.get("max_tokens", dsp.settings.lm.kwargs["max_tokens"])
            max_tokens = min(max(75, max_tokens // 2), max_tokens)
            new_kwargs = {**kwargs, "max_tokens": max_tokens, "n": 1, "temperature": 0.0,}

            assert max_depth > 0
            return generate(template, **new_kwargs)(completion, stage=stage,
                                                    max_depth=max_depth - 1,
                                                    original_example=original_example,)

        completions = Completions(completions, template=template)
        example = example.copy(completions=completions)

        # if len(completions) == 1:
        #     completion = completions[0]
        #     example[stage] = example.copy(**completion)

        #     if dsp.settings.compiling:
        #         inputs_ = set(original_example.keys())
        #         inputs = [
        #             f.input_variable
        #             for f in template.fields
        #             if f.input_variable in inputs_
        #         ]
        #         outputs = [
        #             f.output_variable
        #             for f in template.fields
        #             if f.input_variable not in inputs_
        #         ]

        #         example.compiling_stages = example.get("compiling_stages", [])
        #         example.compiling_stages.append(
        #             {
        #                 "name": stage,
        #                 "template": template,
        #                 "inputs": inputs,
        #                 "outputs": outputs,
        #             },
        #         )
        # else:
        #     # assert not dsp.settings.compiling, "TODO: At this point, cannot compile n>1 generations"
        #     example[stage] = dotdict(completions=completions)

        return example, completions

    return do_generate


# def generate_sc(
#     example, prompt, normalize=True, extract=None, prediction_field=None, **kwargs,
# ):
#     if not dsp.settings.lm:
#         raise AssertionError("No LM is loaded.")
#     kwargs = {"temperature": 0.7, "n": 20, "max_tokens": 150, **kwargs}

#     completions = dsp.settings.lm(prompt, **kwargs)
#     completions = extract_final_answer(example, completions, extract=extract)
#     return majority_vote_(
#         completions, normalize=normalize, prediction_field=prediction_field,
#     )


# def extract_final_answer(example, completions, extract=None):
#     if not dsp.settings.lm:
#         raise AssertionError("No LM is loaded.")
#     if extract:
#         completions = [extract(example, p) for p in completions]
#     else:
#         completions = [
#             p.strip().split("\n")[-1].split(":", 1)[-1].strip() for p in completions
#         ]

#     # TODO: make thread-safe?
#     dsp.settings.lm.history.append(
#         {**dsp.settings.lm.history[-1], "completions": completions},
#     )

#     return completions


# def majority(
#     completions: Completions, normalize: bool = True, field: Optional[str] = None,
# ):
#     """Returns the most common completion for the target field or the last field in the template."""
#     field = completions.template.fields[-1].output_variable if field is None else field

#     return Completions(
#         majority_vote_(completions, normalize=normalize, prediction_field=field),
#         template=completions.template,
#     )


# def majority_vote_(completions: Completions, normalize: bool, prediction_field: str):
#     """Core logic for majority vote."""

#     if not dsp.settings.lm:
#         raise AssertionError("No LM is loaded.")

#     normalized_to_original = {}
#     if normalize:
#         original_completions = completions
#         completions_ = []
#         for pred in completions:
#             if prediction_field in pred:
#                 completions_.append(normalize_text(pred[prediction_field]))
#             else:
#                 completions_.append("")
#         completions = completions_

#         for completion, normalized_completion in zip(original_completions, completions):
#             if normalized_completion not in normalized_to_original:
#                 normalized_to_original[normalized_completion] = completion

#     completions_ = [x for x in completions if x]

#     if completions_:
#         completions = completions_

#     topk = Counter(completions).most_common()
#     pred, _ = topk[0]

#     if normalize:
#         pred = normalized_to_original[pred]

#     dsp.settings.lm.history.append(
#         {**dsp.settings.lm.history[-1], "topk": topk, "completions": [pred]},
#     )

#     return [pred]



"""Retriever model for faiss: https://github.com/facebookresearch/faiss.
Author: Jagane Sundar: https://github.com/jagane.
"""

from typing import Optional, Union

import numpy as np

import dspy
from dsp.modules.sentence_vectorizer import SentenceTransformersVectorizer
from dsp.utils import dotdict

try:
    import faiss
except ImportError:
    faiss = None

if faiss is None:
    raise ImportError(
        """
        The faiss package is required. Install it using `pip install dspy-ai[faiss-cpu]`
        """,
    )


class FaissRM(dspy.Retrieve):
    """A retrieval module that uses an in-memory Faiss to return the top passages for a given query.

    Args:
        document_chunks: the input text chunks
        vectorizer: an object that is a subclass of BaseSentenceVectorizer
        k (int, optional): The number of top passages to retrieve. Defaults to 3.

    Returns:
        dspy.Prediction: An object containing the retrieved passages.

    Examples:
        Below is a code snippet that shows how to use this as the default retriver:
        ```python
        import dspy
        from dspy.retrieve import faiss_rm

        document_chunks = [
            "The superbowl this year was played between the San Francisco 49ers and the Kanasas City Chiefs",
            "Pop corn is often served in a bowl",
            "The Rice Bowl is a Chinese Restaurant located in the city of Tucson, Arizona",
            "Mars is the fourth planet in the Solar System",
            "An aquarium is a place where children can learn about marine life",
            "The capital of the United States is Washington, D.C",
            "Rock and Roll musicians are honored by being inducted in the Rock and Roll Hall of Fame",
            "Music albums were published on Long Play Records in the 70s and 80s",
            "Sichuan cuisine is a spicy cuisine from central China",
            "The interest rates for mortgages is considered to be very high in 2024",
        ]

        frm = faiss_rm.FaissRM(document_chunks)
        turbo = dspy.OpenAI(model="gpt-3.5-turbo")
        dspy.settings.configure(lm=turbo, rm=frm)
        print(frm(["I am in the mood for Chinese food"]))
        ```

        Below is a code snippet that shows how to use this in the forward() function of a module
        ```python
        self.retrieve = FaissRM(k=num_passages)
        ```
    """

    def __init__(self, document_chunks, vectorizer=None, k: int = 3):
        """Inits the faiss retriever.

        Args:
            document_chunks: a list of input strings.
            vectorizer: an object that is a subclass of BaseTransformersVectorizer.
            k: number of matches to return.
        """
        if vectorizer:
            self._vectorizer = vectorizer
        else:
            self._vectorizer = SentenceTransformersVectorizer()
        embeddings = self._vectorizer(document_chunks)
        xb = np.array(embeddings)
        d = len(xb[0])
        dspy.logger.info(f"FaissRM: embedding size={d}")
        if len(xb) < 100:
            self._faiss_index = faiss.IndexFlatL2(d)
            self._faiss_index.add(xb)
        else:
            # if we have at least 100 vectors, we use Voronoi cells
            nlist = 100
            quantizer = faiss.IndexFlatL2(d)
            self._faiss_index = faiss.IndexIVFFlat(quantizer, d, nlist)
            self._faiss_index.train(xb)
            self._faiss_index.add(xb)

        dspy.logger.info(f"{self._faiss_index.ntotal} vectors in faiss index")
        self._document_chunks = document_chunks  # save the input document chunks

        super().__init__(k=k)

    def _dump_raw_results(self, queries, index_list, distance_list) -> None:
        for i in range(len(queries)):
            indices = index_list[i]
            distances = distance_list[i]
            dspy.logger.debug(f"Query: {queries[i]}")
            for j in range(len(indices)):
                dspy.logger.debug(f"    Hit {j} = {indices[j]}/{distances[j]}: {self._document_chunks[indices[j]]}")
        return

    def forward(self, query_or_queries: Union[str, list[str]], k: Optional[int] = None, **kwargs) -> dspy.Prediction:
        """Search the faiss index for k or self.k top passages for query.

        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.

        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        queries = [query_or_queries] if isinstance(query_or_queries, str) else query_or_queries
        queries = [q for q in queries if q]  # Filter empty queries
        embeddings = self._vectorizer(queries)
        emb_npa = np.array(embeddings)
        # For single query, just look up the top k passages
        if len(queries) == 1:
            distance_list, index_list = self._faiss_index.search(emb_npa, k or self.k)
            # self._dump_raw_results(queries, index_list, distance_list)
            passages = [(self._document_chunks[ind], ind) for ind in index_list[0]]
            return [dotdict({"long_text": passage[0], "index": passage[1]}) for passage in passages]

        distance_list, index_list = self._faiss_index.search(emb_npa, (k or self.k) * 3, **kwargs)
        # self._dump_raw_results(queries, index_list, distance_list)
        passage_scores = {}
        for emb in range(len(embeddings)):
            indices = index_list[emb]  # indices of neighbors for embeddings[emb] - this is an array of k*3 integers
            distances = distance_list[
                emb
            ]  # distances of neighbors for embeddings[emb] - this is an array of k*3 floating point numbers
            for res in range((k or self.k) * 3):
                neighbor = indices[res]
                distance = distances[res]
                if neighbor in passage_scores:
                    passage_scores[neighbor].append(distance)
                else:
                    passage_scores[neighbor] = [distance]
        # Note re. sorting:
        # first degree sort: number of queries that got a hit with any particular document chunk. More
        # is a better match. This is len(queries)-len(x[1])
        # second degree sort: sum of the distances of each hit returned by faiss. Smaller distance is a better match
        sorted_passages = sorted(passage_scores.items(), key=lambda x: (len(queries) - len(x[1]), sum(x[1])))[
            : k or self.k
        ]
        return [
            dotdict({"long_text": self._document_chunks[passage_index], "index": passage_index})
            for passage_index, _ in sorted_passages
        ]
